/*////////////////////////////////////////////////////////////////////////
   Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

   Permission to use, copy, modify, and distribute this material for any
   purpose and without fee is hereby granted, provided that the above
   copyright notice and this permission notice appear in all copies, and
   that the name of ETL not be used in advertising or publicity pertaining
   to this material without the specific, prior written permission of an
   authorized representative of ETL.
   ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
   MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
   OR IMPLIED WARRANTIES.
   /////////////////////////////////////////////////////////////////////////
   Content-Type:        program/C; charset=US-ASCII
   Program:     ciiserv.c (CII font server/client)
   Author:              Yutaka Sato <ysato@etl.go.jp>
   Description:
   History:
   951001       created
   /////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <string.h>
#include "cii.h"
extern char *strdup ();

#define A_7bit	"7bit"
#define A_8bit	"8bit"
#define A_1byte	"1byte"
#define A_2byte	"2byte"

#define DELIMITER "--\r\n"

extern struct CharCodesXBM CII_CC[];
extern int CII_CC_SIZE;

CII_lsFonts (buff, size)
     char *buff;
{
  int ci;
  char *bp;
  struct CharCodesXBM *csx;

  bp = buff;
  for (ci = 0; ci < CII_CC_SIZE; ci++)
    {
      csx = &CII_CC[ci];
      if (csx->name == NULL)
	break;
      sprintf (bp, "%-10s %s/%dx%d\n", csx->name, csx->regname,
	       csx->width, csx->height);
      bp += strlen (bp);
    }
  return ci;
}

static struct CharCodesXBM *
CharSetXBM (name)
     char *name;
{
  int ci;
  struct CharCodesXBM *csx;
  char fname[128];
  int width, height;

  fname[0] = 0;
  width = height = 0;
  sscanf (name, "%[^/]/%dx%d", fname, &width, &height);

  for (ci = 0; ci < CII_CC_SIZE; ci++)
    {
      csx = &CII_CC[ci];
      if (csx->name == NULL)
	break;
      if (strcasecmp (fname, CII_CC[ci].name) == 0
	  || strcasecmp (fname, CII_CC[ci].regname) == 0)
	{
	  if (width == 0 || width == csx->width)
	    if (height == 0 || height == csx->height)
	      return csx;
	}
    }
  return csx;
}
CII_withCharset (charset)
     char *charset;
{
  struct CharCodesXBM *csx;

  csx = CharSetXBM (charset);
  return csx->name != NULL;
}

static char *
Sprintf (out, fmt, a, b, c, d, e, f, g)
     char *out, *fmt, *a, *b, *c, *d, *e, *f, *g;
{
  sprintf (out, fmt, a, b, c, d, e, f, g);
  return out + strlen (out);
}

CII_putFont1 (charset, tc)
     char *charset;
     FILE *tc;
{
  int leng;
  struct CharCodesXBM *csx;
  char head[2048], *hp;
  int map, mapx;
  char *bp;
  int nx, xi, cx, nb, bsize, bi;

  csx = CharSetXBM (charset);
  if (csx->name == 0)
    return 0;

  hp = head;
  hp = Sprintf (hp, "CII-Version: 0.1\n");
  hp = Sprintf (hp, "Name: %s,%s\n", csx->regname, csx->name);
  hp = Sprintf (hp, "Size: %d,%d\n", csx->width, csx->height);
  hp = Sprintf (hp, "Range: %d,%d\n", csx->first, csx->last);
  hp = Sprintf (hp, "Attrs: ");
  if (csx->is2bytes)
    hp = Sprintf (hp, A_2byte);
  else
    hp = Sprintf (hp, A_1byte);
  hp = Sprintf (hp, ",");
  if (csx->has8bit)
    hp = Sprintf (hp, A_8bit);
  else
    hp = Sprintf (hp, A_7bit);
  hp = Sprintf (hp, "\n");
  hp = Sprintf (hp, "\f\n\n");
  fputs (head, tc);
  leng = strlen (head);

  nx = csx->last - csx->first + 1;
  map = 0;
  mapx = 0;
  nb = 0;
  for (xi = 0; xi < nx; xi++)
    {
      if (0 <= csx->xbmind[xi])
	{
	  nb++;
	  map |= (1 << mapx);
	}
      if (++mapx == 8)
	{
	  putc (map, tc);
	  leng++;
	  map = 0;
	  mapx = 0;
	}
    }
  if (mapx != 0)
    {
      putc (map, tc);
      leng++;
    }
  bsize = (csx->width * csx->height) / 8;
  nb = 0;
  for (xi = 0; xi < nx; xi++)
    {
      if ((cx = csx->xbmind[xi]) < 0)
	continue;
      bp = csx->xbmsrc[cx];
      for (bi = 0; bi < bsize; bi++)
	putc (bp[bi], tc);
      nb++;
    }
  leng += nb * bsize;
  return leng;
}

CII_getFont1 (charset, fs)
     char *charset;
     FILE *fs;
{
  struct CharCodesXBM *csx,csx0;
  char head1[1024], name[1024], value[1024];
  char rname[1024], alias[1024];
  char val1[256], val2[256];
  int leng;
  int nx, xi, map, mapx, cx, bytes, nb, bsize, rcc;
  char *xbmsrc;

  csx = &csx0;
  bzero(&csx0,sizeof(csx0));

  leng = 0;
  rname[0] = alias[0] = 0;
  while (fgets (head1, sizeof (head1), fs) != NULL)
    {
      leng += strlen (head1);
      if (head1[0] == '\r' || head1[0] == '\n')
	break;
      if (sscanf (head1, "%[^:]: %s", name, value) != 2)
	continue;
      val1[0] = val2[0] = 0;
      sscanf (value, "%[^,],%s", val1, val2);

      if (strcasecmp (name, "Name") == 0)
	{
	  strcpy (alias, val2);
	  strcpy (rname, val1);
	}
      else if (strcasecmp (name, "Size") == 0)
	{
	  csx->width = atoi (val1);
	  csx->height = atoi (val2);
	}
      else if (strcasecmp (name, "Range") == 0)
	{
	  csx->first = atoi (val1);
	  csx->last = atoi (val2);
	}
      else if (strcasecmp (name, "Attrs") == 0)
	{
	  csx->is2bytes = strstr (value, A_2byte) != 0;
	  csx->has8bit = strstr (value, A_8bit) != 0;
	}
    }
  if (rname[0] == 0 || alias[0] == 0)
    return leng;

  csx = CharSetXBM (rname);
  *csx = csx0;

  csx->name = strdup (alias);
  csx->regname = strdup (rname);

  nx = csx->last - csx->first + 1;
  nb = 0;
  csx->xbmind = (int *) malloc (nx * sizeof (int));
  for (xi = 0; xi < nx; xi += 8)
    {
      map = getc (fs);
      leng++;
      for (mapx = 0; mapx < 8 && xi + mapx < nx; mapx++)
	{
	  if (map & (1 << mapx))
	    {
	      csx->xbmind[xi + mapx] = nb;
	      nb++;
	    }
	  else
	    csx->xbmind[xi + mapx] = -1;
	}
    }
  bsize = (csx->width * csx->height) / 8;
  xbmsrc = (char *) malloc (nb * bsize);
  rcc = fread (xbmsrc, 1, nb * bsize, fs);
  leng += rcc;

  bytes = (csx->width * csx->height) / 8;
  csx->xbmsrc = (char **) malloc (sizeof (char *) * nb);
  for (xi = 0; xi < nb; xi++)
    csx->xbmsrc[xi] = &xbmsrc[xi * bytes];

  return leng;
}

CII_getFonts (charset, fs)
     char *charset;
     FILE *fs;
{
  int leng = 0;
  char delimiter[8];

  for (;;){
    leng += CII_getFont1 (charset, fs);
    if (feof(fs))
      break;
    fgets(delimiter, sizeof(delimiter), fs);
    leng += strlen(delimiter);
    if (strcmp(delimiter, DELIMITER) != 0)
      break;
  }
  return leng;
}

CII_putFonts (charset, fs)
     char *charset;
     FILE *fs;
{
  int leng,leng1;
  int ci;
  char charset1[128];
  struct CharCodesXBM *csx,*csx1;

  csx = CharSetXBM (charset);
  if (csx->name == 0)
    return 0;

  leng = CII_putFont1 (charset, fs);
  fprintf (stderr, "%s: %d (%s)\n", csx->regname, leng, csx->name);
  if (!csx->is2bytes)
  {
    for (ci= 0; ci <= CII_CC_SIZE; ci++)
      {
        csx1 = &CII_CC[ci];
        if (csx1->regname == 0)
          break;
        if (csx != csx1 && !csx1->is2bytes)
          {
            fputs (DELIMITER, fs);
            leng += strlen(DELIMITER);
            leng += leng1 = CII_putFont1 (csx1->regname, fs);
            fprintf (stderr, "%s: %d (%s)\n", csx1->regname, leng1, csx1->name);
          }
      }
  }
  fprintf (stderr, "total: %d\n", leng);
  return leng;
}


#ifdef MAIN
main (ac, av)
     char *av[];
{
  int ai, leng;
  char *charset;
  FILE *out = stdout;

  if (ac <= 1)
    {
       fprintf (stderr,"Usage: %s font1 font2 ...\n", av[0]);
       exit (1);
    }
  for (ai = 1; ai < ac; ai++)
    {
      charset = av[ai];
      leng = CII_putFonts (charset, out);
      if (leng <= 0){
        if (1 < ac)
          fputs (DELIMITER, out);
        fprintf (stderr, "%s: not found.\n", charset);
        exit (1);
      }
    }
  exit (0);
}
#endif
